/*
 * Skeleton-code behorende bij het college Netwerken, opleiding Informatica,
 * Universiteit Leiden.
 */

#include "packetbuffer.h"

#include "common.h"

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>

#include <assert.h>

#define MAX(a, b)  ((a) < (b) ? (b) : (a))

/* Initializes the packet buffer. */
void
packet_buffer_init(struct packet_buffer_t *packet_buffer)
{
  pthread_mutex_init(&packet_buffer->mutex, NULL);
  pthread_cond_init(&packet_buffer->cond_taken, NULL);
  pthread_cond_init(&packet_buffer->cond_put, NULL);

  packet_buffer->max_utilization = 0;
  packet_buffer->n_packets = 0;
  packet_buffer->in_pos = 0;
  packet_buffer->out_pos = 0;
}

/* Returns "true" if the packet buffer is empty, "false" otherwise. */
bool
packet_buffer_empty(struct packet_buffer_t *packet_buffer)
{
  bool ret = true;

  pthread_mutex_lock(&packet_buffer->mutex);
  ret = packet_buffer->n_packets > 0 ? false : true;
  pthread_mutex_unlock(&packet_buffer->mutex);

  return ret;
}

/* Puts "packet" on the packet buffer. "drop_perc" is the probability %
 * to use to determine whether or not this packet will be dropped.
 */
void
packet_buffer_put(struct packet_buffer_t *packet_buffer,
                  struct packet_t *packet,
                  const float drop_perc)
{
  if (drop_perc != 0.0 && packet_drop(drop_perc))
    {
#ifdef DEBUG
      puts("(packet has been dropped)");
#endif
      return;
    }

  pthread_mutex_lock(&packet_buffer->mutex);

  while (packet_buffer->n_packets >= PACKET_BUFFER_SIZE)
    pthread_cond_wait(&packet_buffer->cond_taken, &packet_buffer->mutex);

  memcpy(&packet_buffer->buffer[packet_buffer->in_pos], packet,
         sizeof(struct packet_t));

  packet_buffer->in_pos = (packet_buffer->in_pos + 1) % PACKET_BUFFER_SIZE;
  packet_buffer->n_packets++;
  packet_buffer->max_utilization =
      MAX(packet_buffer->n_packets, packet_buffer->max_utilization);

  pthread_cond_signal(&packet_buffer->cond_put);
  pthread_mutex_unlock(&packet_buffer->mutex);
}

static inline void
compute_absolute_time(const long wait_time_msecs,
                      struct timespec *ts)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  ts->tv_sec = tv.tv_sec;
  ts->tv_nsec = tv.tv_usec * 1000L;

  ts->tv_sec += wait_time_msecs / 1000;
  ts->tv_nsec += (wait_time_msecs % 1000) * 1000L * 1000L;
  if (ts->tv_nsec >= 1000000000L)
    {
      ts->tv_sec += 1;
      ts->tv_nsec -= 1000000000L;
    }
}

/* Gets a packet from the packet buffer, the packet will be returned in the
 * "packet" pointer. The return value is true if a packet has been received,
 * false otherwise. If "timeout_msecs" is zero, this function will block
 * until a packet has been received. Otherwise, it will wait for up to
 * "timeout_msecs" milliseconds, returning false if not packet was
 * received within this time.
 */
bool
packet_buffer_get(struct packet_buffer_t *packet_buffer,
                  struct packet_t *packet,
                  const long timeout_msecs)
{
  pthread_mutex_lock(&packet_buffer->mutex);

  while (packet_buffer->n_packets <= 0)
    {
      if (timeout_msecs == 0)
        pthread_cond_wait(&packet_buffer->cond_put,
                          &packet_buffer->mutex);
      else
        {
          struct timespec ts;
          compute_absolute_time(timeout_msecs, &ts);

          int ret = pthread_cond_timedwait(&packet_buffer->cond_put,
                                           &packet_buffer->mutex,
                                           &ts);
          if (ret == ETIMEDOUT)
            {
              assert(packet_buffer->n_packets == 0);
              pthread_mutex_unlock(&packet_buffer->mutex);
              return false;
            }
        }
    }

  memcpy(packet, &packet_buffer->buffer[packet_buffer->out_pos],
         sizeof(struct packet_t));

  packet_buffer->out_pos = (packet_buffer->out_pos + 1) % PACKET_BUFFER_SIZE;
  packet_buffer->n_packets--;

  pthread_cond_signal(&packet_buffer->cond_taken);
  pthread_mutex_unlock(&packet_buffer->mutex);

  return true;
}
